import os
import pandas as pd
from typing import List, Tuple, Dict
from ortools.constraint_solver import routing_enums_pb2, pywrapcp
from geopy.distance import great_circle

class ORTSolver:
    
    def __init__(self,
                 df: pd.DataFrame,
                 start: Tuple[int, List[float]],
                 end: Tuple[int, List[float]],
                 scaling_factor: int = 100,
                 vehicles: int = 1
                 ) -> None:
        self.time_limit = 2
        self.df = df
        self.start_node_id = start[0]  # Start Pillar box ID
        self.end_node_id = end[0]      # End Pillar box ID
        self.start_coords = start[1]   # Start coordinates
        self.end_coords = end[1]       # End coordinates
        self.scaling_factor = scaling_factor
        self.vehicles = vehicles
        self.data_model = self.create_data_model()

    def create_data_model(self) -> Dict:
        """Stores the data for the problem."""
        
        print(f'CREATING DATA MODEL FOR {len(self.df)} BOXES')
        
        # Extract necessary columns from the DataFrame: Pillar box IDs, latitude, longitude
        self.df = self.df[['Pillar box ID', 'latitude', 'longitude']]
        
        # Combine the start and end nodes into the DataFrame if they aren't present
        if self.start_node_id not in self.df['Pillar box ID'].values:
            start_entry = pd.DataFrame({'Pillar box ID': [self.start_node_id], 'latitude': [self.start_coords[0]], 'longitude': [self.start_coords[1]]})
            self.df = pd.concat([start_entry, self.df], ignore_index=True)
        
        if self.end_node_id not in self.df['Pillar box ID'].values:
            end_entry = pd.DataFrame({'Pillar box ID': [self.end_node_id], 'latitude': [self.end_coords[0]], 'longitude': [self.end_coords[1]]})
            self.df = pd.concat([self.df, end_entry], ignore_index=True)

        # Create a mapping dictionary to map Pillar box IDs to indices in the cost matrix
        node_to_index = {node_id: idx for idx, node_id in enumerate(self.df['Pillar box ID'])}
        
        # Initialize the cost matrix
        num_nodes = len(self.df)
        cost_matrix = [[0] * num_nodes for _ in range(num_nodes)]
        
        # Create the cost matrix using great-circle distances 
        # (basicaly haversine, best to use an API for an actual distance matrix)
        coordinates = self.df[['latitude', 'longitude']].values
        for i in range(num_nodes):
            for j in range(num_nodes):
                if i != j:
                    start_coords = coordinates[i]
                    end_coords = coordinates[j]
                    # Calculate the distance (in km) and scale it to an integer
                    cost_matrix[i][j] = int(great_circle(start_coords, end_coords).km * self.scaling_factor)
        
        print(f'VERTEX DATAFRAME CONVERTED TO COST MATRIX OF SIZE {num_nodes}x{num_nodes}')
        
        # Prepare the data dictionary with the cost matrix and start/end nodes
        data = {
            'cost_matrix': cost_matrix,
            'starts': [node_to_index[self.start_node_id]] * self.vehicles,
            'ends': [node_to_index[self.end_node_id]] * self.vehicles,
            'num_vehicles': self.vehicles
        }
        print(data['starts'], data['ends'])
        print(f'DATA MODEL CREATED WITH {len(self.df)} NODES')
        
        return data

    def get_routes(self, solution, routing, manager):
        """
        Get vehicle routes and store them in a two dimensional array whose
        i,j entry is the jth location visited by vehicle i along its route.
        """
        
        routes = []
        for route_nbr in range(routing.vehicles()):
            index = routing.Start(route_nbr)
            route = [manager.IndexToNode(index)]
            while not routing.IsEnd(index):
                index = solution.Value(routing.NextVar(index))
                route.append(manager.IndexToNode(index))
            routes.append(route)
        return routes

    def cost_callback(self,from_index, to_index):
        """Returns the cost between the two nodes."""
        # Convert from routing variable Index to cost matrix NodeIndex.
        from_node = self.manager.IndexToNode(from_index)
        to_node = self.manager.IndexToNode(to_index)
        
        return self.data_model["cost_matrix"][from_node][to_node]

    def print_route_in_index_form(self, vehicle_id: int, route: List[int]):
        """Print the route in index form for a given vehicle."""
        print(f"Route in index form for driver {vehicle_id}: {route}")

    def print_route_in_box_id_form(self, vehicle_id: int, route: List[int]):
        """Print the route in box ID form for a given vehicle."""
        route_box_ids = [int(self.df.iloc[idx]['Pillar box ID']) for idx in route]  # Convert to Python int
        print(f"Route in box ID form for driver {vehicle_id}: {route_box_ids}")

    def calculate_route_cost(self, route: List[int]) -> float:
        """Calculate the total cost of the given route."""
        total_cost = 0
        for i in range(len(route) - 1):
            from_node = route[i]
            to_node = route[i + 1]
            total_cost += self.data_model['cost_matrix'][from_node][to_node]
        return total_cost

    def solve(self):
        
        # Create the routing index manager.
        self.manager = pywrapcp.RoutingIndexManager(len(self.data_model['cost_matrix']),
                                            self.vehicles,
                                            self.data_model['starts'],
                                            self.data_model['ends']
                                            )
        # Create Routing Model.
        routing_model = pywrapcp.RoutingModel(self.manager)
        # Create and register a transit callback.
        transit_callback_index = routing_model.RegisterTransitCallback(self.cost_callback)
        # Define cost of each arc.
        routing_model.SetArcCostEvaluatorOfAllVehicles(transit_callback_index)
        '''
        Add any constraints here. This is best configured to your specific use case.
        Recommendation:  to make manager, routing_model, and transit_callback_index class attributes.
        then you can define a function to add constraints so that you can do it dynamically on any solver instance.
        '''
        # ...
        # Cost here refers to distance since we use haversine in the cost matrix
        dimension_name = 'Cost'
        routing_model.AddDimension(
            transit_callback_index,
            0,  # no slack
            3000,  # maximum cost per vehicle
            True,  # start cumul to zero
            dimension_name)
        distance_dimension = routing_model.GetDimensionOrDie(dimension_name)
        distance_dimension.SetGlobalSpanCostCoefficient(100)
        
        '''
        All of these solver configurations can be converted to class attributes 
        to be set in the constructor
        '''
        # Setting first solution heuristic
        search_parameters = pywrapcp.DefaultRoutingSearchParameters()
        # Put an upper limit on the time taken by the solver
        search_parameters.time_limit.seconds = self.time_limit
        # We chose SAVINGS as our first solution strategy. Check ORTools documentation for more options.
        search_parameters.first_solution_strategy = (
            routing_enums_pb2.FirstSolutionStrategy.SAVINGS)
        # You can also define metaheuristics here. Check ORTools documentation for more options.
        #...
        search_parameters.local_search_metaheuristic = (
            routing_enums_pb2.LocalSearchMetaheuristic.GUIDED_LOCAL_SEARCH)
        
        # Run the problem solver
        solution = routing_model.SolveWithParameters(search_parameters)
       
        if solution:
            # Initialize result structure
            result = {'routes': [], 'costs': []}

            # Get the routes for each vehicle
            routes = self.get_routes(solution, routing_model, self.manager)

            for vehicle_id, route in enumerate(routes):
                print(f"\n--- Route for driver {vehicle_id} ---")

                # Print routes in index and box ID form
                self.print_route_in_index_form(vehicle_id, route)
                self.print_route_in_box_id_form(vehicle_id, route)

                # Get the total cost for this route (scaled)
                route_cost = self.calculate_route_cost(route)
                adjusted_cost = route_cost / self.scaling_factor
                print(f"Total cost for driver {vehicle_id}'s route (rescaled): {adjusted_cost}")

                # Save route and cost in result structure
                route_box_ids = [int(self.df.iloc[idx]['Pillar box ID']) for idx in route]  # Cast to int
                result['routes'].append({'driver': vehicle_id, 'route_ids': route_box_ids})
                result['costs'].append({'driver': vehicle_id, 'cost': float(adjusted_cost)})  # Cast to float

            return result
        else:
            print("No solution found.")
            return None
            

